@@ -1,5 +1,5 @@
 /*
- * $Id: mod_ruby.c,v 1.86 2004/04/30 05:44:18 shugo Exp $
+ * $Id: mod_ruby.c 37 2004-09-25 03:59:57Z shugo $
  * Copyright (C) 2000  ZetaBITS, Inc.
  * Copyright (C) 2000  Information-technology Promotion Agency, Japan
  * Copyright (C) 2001  Shugo Maeda <shugo@modruby.net>
@@ -63,16 +63,29 @@
 #include <mach-o/dyld.h>
 #endif
 
-#ifndef WIN32
+#ifdef _WIN32
+#define GET_ENVIRON(e) (e = rb_w32_get_environ())
+#define FREE_ENVIRON(e) rb_w32_free_environ(e)
+static char **my_environ;
+#undef environ
+#define environ my_environ
+#elif defined(__APPLE__)
+#undef environ
+#define environ (*_NSGetEnviron())
+#define GET_ENVIRON(e) (e)
+#define FREE_ENVIRON(e)
+#else
 extern char **environ;
-static char **origenviron;
-#endif /* WIN32 */
+#define GET_ENVIRON(e) (e)
+#define FREE_ENVIRON(e)
+#endif
 
-EXTERN VALUE ruby_errinfo;
-EXTERN VALUE rb_stdin;
-EXTERN VALUE rb_stdout;
+RUBY_EXTERN VALUE ruby_errinfo;
+RUBY_EXTERN VALUE rb_stdin;
+RUBY_EXTERN VALUE rb_stdout;
+RUBY_EXTERN VALUE rb_stderr;
 #if RUBY_VERSION_CODE < 180
-EXTERN VALUE rb_defout;
+RUBY_EXTERN VALUE rb_defout;
 #endif
 
 static VALUE orig_stdin;
@@ -81,7 +94,7 @@ static VALUE orig_stdout;
 static VALUE orig_defout;
 #endif
 
-EXTERN VALUE rb_load_path;
+RUBY_EXTERN VALUE rb_load_path;
 static VALUE default_load_path;
 
 static const char *default_kcode;
@@ -93,9 +106,13 @@ array_header *ruby_required_libraries = NULL;
 #include "apr_thread_cond.h"
 
 static apr_thread_t *ruby_thread;
+static apr_thread_mutex_t *ruby_is_running_mutex;
+static apr_thread_cond_t *ruby_is_running_cond;
 static apr_thread_mutex_t *ruby_request_queue_mutex;
 static apr_thread_cond_t *ruby_request_queue_cond;
 
+int ruby_is_threaded_mpm;
+
 typedef VALUE (*ruby_protect_func_t)(VALUE);
 
 typedef struct ruby_request {
@@ -113,6 +130,8 @@ static ruby_request_t *ruby_request_queue = NULL;
 #define SHUTDOWN_RUBY_THREAD NULL
 #endif
 
+#define RUBY_OBJECT_HANDLER "ruby-object"
+
 static int ruby_object_handler(request_rec *r);
 static int ruby_trans_handler(request_rec *r);
 static int ruby_authen_handler(request_rec *r);
@@ -172,6 +191,9 @@ static const command_rec ruby_cmds[] =
     {"RubyCleanupHandler", ruby_cmd_cleanup_handler,
      NULL, OR_ALL, TAKE1,
      "set cleanup handler object"},
+    {"RubyChildInitHandler", ruby_cmd_child_init_handler,
+     NULL, RSRC_CONF, TAKE1,
+     "set child init handler object"},
     {NULL}
 };
 
@@ -215,7 +237,7 @@ static void ruby_child_init(server_rec*, pool*);
 
 static const handler_rec ruby_handlers[] =
 {
-    {"ruby-object", ruby_object_handler},
+    {RUBY_OBJECT_HANDLER, ruby_object_handler},
     {NULL}
 };
 
@@ -392,13 +414,13 @@ static void get_exception_info(VALUE str)
 	len = ep->len;
 	for (i=1; i<len; i++) {
 	    if (TYPE(ep->ptr[i]) == T_STRING) {
-		STR_CAT_LITERAL(str, "\tfrom ");
+		STR_CAT_LITERAL(str, "  from ");
 		rb_str_cat(str, RSTRING(ep->ptr[i])->ptr, RSTRING(ep->ptr[i])->len);
 		STR_CAT_LITERAL(str, "\n");
 	    }
 	    if (i == TRACE_HEAD && len > TRACE_MAX) {
 		char buff[BUFSIZ];
-		snprintf(buff, BUFSIZ, "\t ... %ld levels...\n",
+		snprintf(buff, BUFSIZ, "   ... %ld levels...\n",
 			 len - TRACE_HEAD - TRACE_TAIL);
 		rb_str_cat(str, buff, strlen(buff));
 		i = len - TRACE_TAIL;
@@ -449,7 +471,7 @@ VALUE ruby_get_error_info(int state)
 }
 
 void ruby_log_error(const char *file, int line, int level,
-			const server_rec *s, const char *fmt, ...)
+                    const server_rec *s, const char *fmt, ...)
 {
     va_list args;
     char buf[BUFSIZ];
@@ -466,12 +488,15 @@ void ruby_log_error(const char *file, int line, int level,
 
 void ruby_log_error_string(server_rec *s, VALUE errmsg)
 {
-    VALUE logmsg;
+    VALUE msgs;
+    int i;
 
-    logmsg = STRING_LITERAL("error in ruby\n");
-    rb_str_concat(logmsg, errmsg);
-    ruby_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, s,
-		   "%s", StringValuePtr(logmsg));
+    ruby_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, s, "error in ruby");
+    msgs = rb_str_split(errmsg, "\n");
+    for (i = 0; i < RARRAY(msgs)->len; i++) {
+        ruby_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, s,
+                       "%s", StringValuePtr(RARRAY(msgs)->ptr[i]));
+    }
 }
 
 static void handle_error(request_rec *r, int state)
@@ -479,9 +504,11 @@ static void handle_error(request_rec *r, int state)
     VALUE errmsg, reqobj;
 
     errmsg = ruby_get_error_info(state);
-    reqobj = (VALUE) ap_get_module_config(r->request_config, &ruby_module);
-    if (reqobj)
-	rb_apache_request_set_error(reqobj, errmsg, ruby_errinfo);
+    if (r->request_config) {
+	reqobj = (VALUE) ap_get_module_config(r->request_config, &ruby_module);
+	if (reqobj)
+	    rb_apache_request_set_error(reqobj, errmsg, ruby_errinfo);
+    }
     ruby_log_error_string(r->server, errmsg);
 }
 
@@ -544,6 +571,17 @@ static void ruby_add_path(const char *path)
 static int ruby_startup(pool *p, pool *plog, pool *ptemp, server_rec *s)
 {
     ap_add_version_component(p, MOD_RUBY_STRING_VERSION);
+#if RUBY_RELEASE_CODE > 20040624
+    {
+        char *version = apr_pstrcat(p, "Ruby/", ruby_version,
+                                    "(", ruby_release_date, ")",
+				    (char *) NULL);
+        ap_add_version_component(p, version);
+    }
+#endif
+#if APR_HAS_THREADS
+    ruby_is_threaded_mpm = ap_find_linked_module("prefork.c") == NULL;
+#endif
     return OK;
 }
 
@@ -552,6 +590,14 @@ static int ruby_startup(pool *p, pool *plog, pool *ptemp, server_rec *s)
 static void ruby_startup(server_rec *s, pool *p)
 {
     ap_add_version_component(MOD_RUBY_STRING_VERSION);
+#if RUBY_RELEASE_CODE > 20040624
+    {
+        char *version = ap_pstrcat(p, "Ruby/", ruby_version,
+                                   "(", ruby_release_date, ")",
+				   (char *) NULL);
+        ap_add_version_component(version);
+    }
+#endif
 }
 
 #endif
@@ -610,15 +656,14 @@ static void ruby_init_interpreter(server_rec *s)
     rb_define_global_const("MOD_RUBY",
 			   STRING_LITERAL(MOD_RUBY_STRING_VERSION));
 
-#ifndef WIN32
-    origenviron = environ;
-#endif /* WIN32 */
-
     orig_stdin = rb_stdin;
     orig_stdout = rb_stdout;
 #if RUBY_VERSION_CODE < 180
     orig_defout = rb_defout;
 #endif
+#ifdef APACHE2
+    rb_protect_funcall(rb_stderr, rb_intern("sync="), NULL, 1, Qtrue);
+#endif
 
     ruby_init_loadpath();
     default_load_path = rb_load_path;
@@ -661,7 +706,7 @@ static void dso_unload(void *handle)
 
 static void ruby_finalize_interpreter()
 {
-    EXTERN VALUE ruby_dln_librefs;
+    RUBY_EXTERN VALUE ruby_dln_librefs;
     int i;
 
     ruby_finalize();
@@ -678,6 +723,11 @@ static void* APR_THREAD_FUNC ruby_thread_start(apr_thread_t *t, void *data)
 
     ruby_init_interpreter(s);
 
+    apr_thread_mutex_lock(ruby_is_running_mutex);
+    ruby_is_running = 1;
+    apr_thread_cond_signal(ruby_is_running_cond);
+    apr_thread_mutex_unlock(ruby_is_running_mutex);
+
     while (1) {
 	apr_thread_mutex_lock(ruby_request_queue_mutex);
 	while (ruby_request_queue == NULL)
@@ -740,103 +790,145 @@ apr_status_t ruby_call_interpreter(pool *p, ruby_interp_func_t func,
 static APR_CLEANUP_RETURN_TYPE ruby_child_cleanup(void *data)
 {
 #if APR_HAS_THREADS
-    pool *p;
-    apr_status_t status;
+    if (ruby_is_threaded_mpm) {
+	pool *p;
+	apr_status_t status, result;
 
-    status = apr_pool_create(&p, NULL);
-    if (status != APR_SUCCESS)
+#ifdef SIGTERM
+	ruby_signal(SIGTERM, SIG_IGN);
+#endif
+	status = apr_pool_create(&p, NULL);
+	if (status != APR_SUCCESS)
+	    return status;
+	status = ruby_call_interpreter(p, SHUTDOWN_RUBY_THREAD, NULL, NULL, NULL);
+	apr_pool_clear(p);
+	if (status != APR_SUCCESS)
+	    return status;
+	status = apr_thread_join(&result, ruby_thread);
 	return status;
-    status = ruby_call_interpreter(p, SHUTDOWN_RUBY_THREAD, NULL, NULL, NULL);
-    apr_pool_clear(p);
-    return status;
-#else
-    ruby_finalize_interpreter();
+    }
+    else {
+#endif
+	ruby_finalize_interpreter();
+	APR_CLEANUP_RETURN_SUCCESS();
+#if APR_HAS_THREADS
+    }
 #endif
-    APR_CLEANUP_RETURN_SUCCESS();
 }
 
+static request_rec *fake_request_rec(server_rec *s, pool *p, char *hook)
+{
+    request_rec *r = (request_rec *) ap_pcalloc(p, sizeof(request_rec));
+    r->pool = p; 
+    r->server = s;
+    r->per_dir_config = NULL;
+    r->request_config = NULL;
+    r->uri = hook;
+    r->notes = NULL;
+    return r;
+}
+
+static int ruby_handler(request_rec *, array_header *, ID, int, int);
+
 #ifdef APACHE2
 static void ruby_child_init(pool *p, server_rec *s)
 #else /* Apache 1.x */
 static void ruby_child_init(server_rec *s, pool *p)
 #endif
 {
+    ruby_server_config *conf;
+    request_rec *r;
+
     if (!ruby_running()) {
 #if APR_HAS_THREADS
-	apr_status_t status;
-	status = apr_thread_mutex_create(&ruby_request_queue_mutex,
-					 APR_THREAD_MUTEX_DEFAULT, p);
-	if (status != APR_SUCCESS) {
-	    ruby_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, s,
-			   "failed to create mutex");
+	if (ruby_is_threaded_mpm) {
+	    apr_status_t status;
+	    status = apr_thread_mutex_create(&ruby_is_running_mutex,
+					     APR_THREAD_MUTEX_DEFAULT, p);
+	    if (status != APR_SUCCESS) {
+		ruby_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, s,
+			       "failed to create mutex");
+		return;
+	    }
+	    status = apr_thread_cond_create(&ruby_is_running_cond, p);
+	    if (status != APR_SUCCESS) {
+		ruby_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, s,
+			       "failed to create cond");
+		return;
+	    }
+	    status = apr_thread_mutex_create(&ruby_request_queue_mutex,
+					     APR_THREAD_MUTEX_DEFAULT, p);
+	    if (status != APR_SUCCESS) {
+		ruby_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, s,
+			       "failed to create mutex");
+		return;
+	    }
+	    status = apr_thread_cond_create(&ruby_request_queue_cond, p);
+	    if (status != APR_SUCCESS) {
+		ruby_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, s,
+			       "failed to create cond");
+		return;
+	    }
+	    status = apr_thread_create(&ruby_thread, NULL, ruby_thread_start, s, p);
+	    if (status != APR_SUCCESS) {
+		ruby_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, s,
+			       "failed to create ruby thread");
+		return;
+	    }
+	    apr_thread_mutex_lock(ruby_is_running_mutex);
+	    while (!ruby_running())
+		apr_thread_cond_wait(ruby_is_running_cond,
+				     ruby_is_running_mutex);
+	    apr_thread_mutex_unlock(ruby_is_running_mutex);
 	}
-	status = apr_thread_cond_create(&ruby_request_queue_cond, p);
-	if (status != APR_SUCCESS) {
-	    ruby_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, s,
-			   "failed to create cond");
+	else {
+#endif
+	    ruby_init_interpreter(s);
+	    ruby_is_running = 1;
+#if APR_HAS_THREADS
 	}
-	status = apr_thread_create(&ruby_thread, NULL, ruby_thread_start, s, p);
-	if (status != APR_SUCCESS) {
-	    ruby_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, s,
-			   "failed to create ruby thread");
-	}
-#else
-	ruby_init_interpreter(s);
 #endif
 	ap_register_cleanup(p, NULL, ruby_child_cleanup, ap_null_cleanup);
-	ruby_is_running = 1;
     }
+
+    r = fake_request_rec(s, p, "RubyChildInitHandler");
+    conf = get_server_config(r->server);
+    ruby_handler(r, conf->ruby_child_init_handler,
+		 rb_intern("child_init"), 0, 0);
 }
 
-static void mod_ruby_clearenv()
+static void mod_ruby_clearenv(pool *p)
 {
-#ifdef WIN32
-    char *orgp, *p;
-
-    orgp = p = GetEnvironmentStrings();
-
-    if (p == NULL)
-	return;
-
-    while (*p) {
-	char buf[1024];
-	char *q;
-
-	strncpy(buf, p, sizeof buf);
-	q = strchr(buf, '=');
-	if (q)
-	    *(q+1) = '\0';
-
-	putenv(buf);
-	p += strlen(p) + 1;
+    char **env;
+    array_header *names = ap_make_array(p, 1, sizeof(char*));
+    int i;
+    
+    env = GET_ENVIRON(environ);
+    while (*env) {
+        char *s = strchr(*env, '=');
+        if (s) {
+            *(char **) ap_push_array(names) = ap_pstrndup(p, *env, s - *env);
+        }
+        env++;
     }
-
-    FreeEnvironmentStrings(orgp);
-#else
-#ifndef __CYGWIN__
-    if (environ == origenviron) {
-	environ = ALLOC_N(char*, 1);
-    }
-    else {
-	char **p;
-
-	for (p = environ; *p; p++) {
-	    if (*p) free(*p);
+    FREE_ENVIRON(environ);
+    for (i = 0; i < names->nelts; i++) {
+	char *name = ((char **) names->elts)[i];
+	char *val = getenv(name);
+	if (val) {
+	    ruby_unsetenv(name);
 	}
-	REALLOC_N(environ, char*, 1);
     }
-    *environ = NULL;
-#endif
-#endif
 }
 
 static void mod_ruby_setenv(const char *name, const char *value)
 {
     if (!name) return;
+
+    ruby_unsetenv(name);
+
     if (value && *value)
 	ruby_setenv(name, value);
-    else
-	ruby_unsetenv(name);
 }
 
 static void setenv_from_table(table *tbl)
@@ -859,7 +951,7 @@ void rb_setup_cgi_env(request_rec *r)
     ruby_server_config *sconf = get_server_config(r->server);
     ruby_dir_config *dconf = get_dir_config(r);
 
-    mod_ruby_clearenv();
+    mod_ruby_clearenv(r->pool);
     ap_add_common_vars(r);
     ap_add_cgi_vars(r);
     setenv_from_table(r->subprocess_env);
@@ -960,7 +1052,7 @@ static void per_request_init(request_rec *r)
     mod_ruby_setup_loadpath(sconf, dconf);
     ruby_debug = Qfalse;
     ruby_verbose = Qfalse;
-    if (dconf->kcode)
+    if (dconf && dconf->kcode)
 	rb_set_kcode(dconf->kcode);
     rb_request = rb_get_request_object(r);
     rb_stdin = rb_stdout = rb_request;
@@ -1003,6 +1095,7 @@ static void per_request_cleanup(request_rec *r, int flush)
 #endif
     }
     rb_set_kcode(default_kcode);
+    rb_gc();
 }
 
 typedef struct handler_0_arg {
@@ -1070,7 +1163,7 @@ static void *ruby_handler_internal(handler_internal_arg_t *iarg)
 
     sconf = get_server_config(r->server);
     dconf = get_dir_config(r);
-    safe_level = dconf->safe_level;
+    safe_level = dconf ? dconf->safe_level : MR_DEFAULT_SAFE_LEVEL;
     handlers = (char **) handlers_arr->elts;
     handlers_len = handlers_arr->nelts;
     iarg->retval = DECLINED;
@@ -1114,13 +1207,14 @@ static int ruby_handler(request_rec *r,
     arg->flush = flush;
     arg->retval = 0;
 #if APR_HAS_THREADS
-    {
+    if (ruby_is_threaded_mpm) {
 	apr_status_t status;
 	char buf[256];
 
-	status = ruby_call_interpreter(r->pool,
-				       (ruby_interp_func_t) ruby_handler_internal,
-				       arg, NULL, 0);
+	status =
+	    ruby_call_interpreter(r->pool,
+				  (ruby_interp_func_t) ruby_handler_internal,
+				  arg, NULL, 0);
 	if (status != APR_SUCCESS) {
 	    apr_strerror(status, buf, sizeof(buf));
 	    ruby_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r->server,
@@ -1128,8 +1222,11 @@ static int ruby_handler(request_rec *r,
 	    return HTTP_INTERNAL_SERVER_ERROR;
 	}
     }
-#else
-    ruby_handler_internal(arg);
+    else {
+#endif
+	ruby_handler_internal(arg);
+#if APR_HAS_THREADS
+    }
 #endif
     return arg->retval;
 }
@@ -1137,9 +1234,14 @@ static int ruby_handler(request_rec *r,
 static int ruby_object_handler(request_rec *r)
 {
     int retval;
-
-    ruby_dir_config *dconf = get_dir_config(r);
+    ruby_dir_config *dconf;
     
+#ifdef APACHE2
+    if (strcmp(r->handler, RUBY_OBJECT_HANDLER) != 0) {
+	return DECLINED;
+    }
+#endif
+    dconf = get_dir_config(r);
     retval = ruby_handler(r, dconf->ruby_handler, rb_intern("handler"), 0, 1);
 #ifdef APACHE2
     if (retval == DECLINED && r->finfo.filetype == APR_DIR)
@@ -1264,3 +1366,5 @@ static int ruby_post_read_request_handler(request_rec *r)
  * tab-width: 8
  * End:
  */
+
+/* vim: set filetype=c ts=8 sw=4 : */
